"""
2D simple integrator of Boussinesq solution for arbitrary traction distribution on a half-space.
Double integral is accelerated by numba.
V.A. Yastrebov, MINES Paris - PSL, CNRS, Evry/Paris, France, 2024.
Assited by Copilot and GPT4.
License: BSD 3 clause
"""
import numpy as np
import matplotlib.pyplot as plt
# Use numba for performance, it is just-in-time (JIT) compiler for functions
import numba
import time

# ----------------------------------------- #
# Numerical integration for stresses        #
# and displacements, assuming constant      #
# pressure/shear distributions per cell     #
# The integration is accelerated by         #
# numba JIT (just-in-time) compiler         #
# ----------------------------------------- #

@numba.njit
def integrate_stress(x0, y0, angle, extent, Ne, depth, Nd, XX, YY, nu, F,T):
    # Finds stress distribution within a plane oriented at `angle` with respect to OX axis and starting at x0,y0: width = `extent` and depth = `depth`
    dx = extent/(Ne)
    X = np.linspace(dx/2, extent-dx/2, Ne)
    dz = depth/(Nd)
    Z = np.linspace(dz/2, depth-dz/2, Nd)
    N = len(XX)

    # Initialize stress arrays
    sigma_x = np.zeros((Ne, Nd))
    sigma_y = np.zeros((Ne, Nd))
    sigma_z = np.zeros((Ne, Nd))
    sigma_xy = np.zeros((Ne, Nd))
    sigma_yz = np.zeros((Ne, Nd))
    sigma_xz = np.zeros((Ne, Nd))

    # Iterate over all grid points
    for i in range(Ne):
        x = x0+X[i]*np.cos(angle)
        y = y0+X[i]*np.sin(angle)
        for j in range(Nd):
            z = Z[j]
            for k in range(N):
                for l in range(N):
                    xp = XX[k,l]
                    yp = YY[k,l]
                    rho = np.sqrt((x-xp)**2 + (y-yp)**2 + z**2)
                    r   = np.sqrt((x-xp)**2 + (y-yp)**2)
                    factor1 = F[k,l] / (2*np.pi)
                    factor2 = (1 - 2*nu) / r**2
                    if r!=0 and rho!=0:
                        sigma_x[i,j] += factor1 * (factor2 * ((1-z/rho)*((x-xp)**2-(y-yp)**2)/r**2 + z*(y-yp)**2/rho**3) - 3*z*(x-xp)**2/rho**5)
                        sigma_y[i,j] += factor1 * (factor2 * ((1-z/rho)*((y-yp)**2-(x-xp)**2)/r**2 + z*(x-xp)**2/rho**3) - 3*z*(y-yp)**2/rho**5)
                        sigma_z[i,j] += -3*factor1 * (z**3/rho**5)
                        sigma_xy[i,j] += factor1 * (factor2 * ((1-z/rho)*(x-xp)*(y-yp)/r**2 - z*(x-xp)*(y-yp)/rho**3) - 3*z*(x-xp)*(y-yp)/rho**5)
                        sigma_xz[i,j] += -3*factor1 * (z**2*(x-xp)/rho**5)
                        sigma_yz[i,j] += -3*factor1 * (z**2*(y-yp)/rho**5)
                    # elif rho!=0:
                        factor1 = T[k,l] / (2*np.pi)
                        sigma_x[i,j] += factor1 * (-3*(x-xp)**3/rho**5 + (1 - 2*nu) * ((x-xp)/rho**3 - 3*(x-xp)/(rho*(rho+z)**2) + (x-xp)**3/(rho**3*(rho   +z)**2) + 2*(x-xp)**3/(rho**2*(rho+z)**3) ))
                        sigma_y[i,j] += factor1 * (-3*(x-xp)*(y-yp)**2/rho**5 + (1 - 2*nu) * ((x-xp)/rho**3 - (x-xp)/(rho*(rho+z)**2) + (x-xp)*(y-yp)**2/(rho**3*(rho+z)**2) + 2*(x-xp)*(y-yp)**2/(rho**2*(rho+z)**3) )) 
                        sigma_z[i,j] += -3*factor1 * (x-xp)*z**2/rho**5
                        sigma_xy[i,j] += factor1 * (-3*(x-xp)**2*(y-yp)/rho**5 + (1 - 2*nu) * (-(y-yp)/(rho*(rho+z)**2) + (x-xp)**2*(y-yp)/(rho**3*(rho+z)**2) + 2*(x-xp)**2*(y-yp)/(rho**2*(rho+z)**3) )) 
                        sigma_xz[i,j] += -3*factor1 * (x-xp)*(y-yp)*z/rho**5
                        sigma_yz[i,j] += -3*factor1 * (x-xp)**2*z/rho**5                        

    return sigma_x, sigma_y, sigma_z, sigma_xy, sigma_yz, sigma_xz


@numba.njit
def integrate_surface_displacement_for_pressure(X, XX, Y, YY, nu, G, F):
    # Initialize the displacement arrays
    Ux = np.zeros((N, N))
    Uy = np.zeros((N, N))
    Uz = np.zeros((N, N))

    # Iterate over all grid points
    for i in range(N):
        for j in range(N):
            for k in range(N):
                for l in range(N):
                    if not (i == k and j == l):  
                        r = np.sqrt((X[i,j]-XX[k,l])**2 + (Y[i,j]-YY[k,l])**2)
                        factor = F[k,l]/(4*np.pi*G)
                        Ux[i,j] += - factor * (1 - 2*nu) * (X[i,j]-XX[k,l])/r**2
                        Uy[i,j] += - factor * (1 - 2*nu) * (Y[i,j]-YY[k,l])/r**2
                        Uz[i,j] += - 2 * factor * (1 - nu) / r
                        pass

    return Ux, Uy, Uz

@numba.njit
def integrate_surface_displacement_for_shear(X, XX, Y, YY, nu, G, T):
    # Initialize the displacement arrays
    Ux = np.zeros((N, N))
    Uy = np.zeros((N, N))
    Uz = np.zeros((N, N))

    # Iterate over all grid points
    for i in range(N):
        for j in range(N):
            for k in range(N):
                for l in range(N):
                    if not (i == k and j == l):  
                        rho = np.sqrt((X[i,j]-XX[k,l])**2 + (Y[i,j]-YY[k,l])**2)
                        factor = T[k,l]/(4*np.pi*G)

                        term1 = (1 - 2*nu) * (1 / rho - (X[i,j]-XX[k,l])**2 / rho**3)
                        Ux[i,j] += factor * (1 / rho - (X[i,j]-XX[k,l])**2 / rho**3 + term1)
                        Uy[i,j] += - factor * (X[i,j]-XX[k,l]) * (Y[i,j]-YY[k,l]) * 2*nu / rho**3
                        Uz[i,j] += factor * (1 - 2*nu) * (X[i,j]-XX[k,l]) / rho**2
                        pass

    return Ux, Uy, Uz

def pressure(x, y, x0, y0, a, p0):
    r2 = (x-x0)**2 + (y-y0)**2
    if r2 <= a**2:        
        return p0*np.sqrt(1-r2/a**2)
    else:
        return 0
Pressure = np.vectorize(pressure)

# ----------------------------------------- #
# Parameters                                #
# ----------------------------------------- #

# Physical parameters
x_min = y_min = 0
x_max = y_max = 0.1
x0 = y0 = (x_max-x_min)/2
a = 0.02

# Material parameters
E = 2.1e11
nu = 0.3
G = E/(2*(1+nu))

# Loading parameters
p0 = 1e3 
friction = 0.

# Computational parameters
N = 100
dx = (x_max - x_min)/N
uz = np.zeros((N,N))
ux = np.zeros((N,N))
uy = np.zeros((N,N))

x = np.linspace(x_min + dx/2, x_max - dx/2, N)
y = np.linspace(y_min + dx/2, y_max - dx/2, N)
X, Y = np.meshgrid(x, y)
XX, YY = np.meshgrid(x, y)

P = Pressure(XX,YY,x0,y0,a,p0)
Q = friction*P
F = P*dx**2
Tx = Q*dx**2
ComputeDisplacement = False
ComputeStress = True

# ----------------------------------------- #
# Plotting pressure / shear distribution    #
# ----------------------------------------- #

fig, ax = plt.subplots(1,2, figsize=(10, 4))  
# Plot pressure
pressure_img = ax[0].imshow(P/p0, extent=[x_min + dx/2, x_max, y_min + dx/2, y_max], 
                            origin='lower', interpolation='bicubic', cmap=plt.cm.inferno, vmax = 1)
fig.colorbar(pressure_img, ax=ax[0], label='Pressure', orientation="horizontal")
ax[0].set_title('Pressure Distribution')
ax[0].set_xlabel('x')
ax[0].set_ylabel('y')
# Plot shear
shear_img = ax[1].imshow(Q/p0, extent=[x_min + dx/2, x_max, y_min + dx/2, y_max], 
                          origin='lower', interpolation='bicubic', cmap=plt.cm.inferno, vmax = 1)
fig.colorbar(shear_img, ax=ax[1], label='Shear', orientation="horizontal")
ax[1].set_title('Shear Distribution')
ax[1].set_xlabel('x')
ax[1].set_ylabel('y')
# Plot circles of radius $a$ on both subplots
theta = np.linspace(0, 2 * np.pi, 100)
ax[0].plot(x0 + a * np.cos(theta), y0 + a * np.sin(theta), linestyle="--", color='grey', zorder=2)
ax[1].plot(x0 + a * np.cos(theta), y0 + a * np.sin(theta), linestyle="--", color='grey', zorder=2)
plt.tight_layout()  
plt.show()

# -------------------------------------------- #
# Compute and plot stresses and displacements  #
# -------------------------------------------- #

if ComputeStress:
    # Define section
    angle = 0                       # Rotation with respect to OX axis
    extent = x_max                  # Length of the section
    depth = 0.2*x_max               # Depth of the section
    Ne = 100                        # Discretization in length
    Nd = 40                         # Discretization in depth
    x0 = 0                          # start point X
    y0 = y_min + (y_max-y_min)/2    # start point in Y

    # Compute stresses due to pressure and shear
    time0 = time.time()
    sigma_x, sigma_y, sigma_z, sigma_xy, sigma_yz, sigma_xz = integrate_stress(x0,y0,angle,extent,Ne,depth,Nd, XX, YY, nu, F, Tx)
    print("Time elapsed: {0:.2f} s".format(time.time()-time0))

    # Plot von Mises stress
    sigma_vm = np.sqrt(0.5*((sigma_x-sigma_y)**2 + (sigma_y-sigma_z)**2 + (sigma_z-sigma_x)**2) + 3*(sigma_xy**2 + sigma_yz**2 + sigma_xz**2))
    fig,ax = plt.subplots()
    if angle == 0:
        ax.set_xlabel('x')
    elif angle == np.pi/2. or angle == -np.pi/2.:
        ax.set_xlabel('y')
    else:
        ax.set_xlabel('$x \cos({0:.2f}) + y \sin({0:.2f})$'.format(angle))
    ax.set_ylabel('z')
    plt.title("von Mises stress")
    field =sigma_vm.T
    bar = ax.imshow(field, extent=[0, extent, -depth, 0], origin='upper',interpolation='bicubic', cmap=plt.cm.coolwarm)
    plt.colorbar(bar, label='von Mises stress', orientation="horizontal")

    dx = extent/(Ne)
    dz = depth/(Nd)
    x = np.linspace(dx/2, extent-dx/2, Ne)
    z = np.linspace(-dz/2,-depth+dz/2, Nd)
    plt.contour(x, z, field, 20, colors='white', linewidths=1, linestyle = "--")
    plt.tight_layout()
    # To save von Mises stress figure, uncomment this line
    # fig.savefig("von_Mises_stress_p0_{0:.3f}_a_{1:.3f}.pdf".format(p0,a))
    plt.show()


if ComputeDisplacement:
    # Compute displacements from the pressure component
    uxn,uyn,uzn = integrate_surface_displacement_for_pressure(X, XX, Y, YY, nu, G, F)
    # Compute displacements from the shear component
    ux,uy,uz = integrate_surface_displacement_for_shear(X, XX, Y, YY, nu, G, Tx)
    # Sum them up
    ux += uxn
    uy += uyn
    uz += uzn

    # Plot vertical displacement
    fig,ax = plt.subplots()
    ax.set_xlabel('x')
    ax.set_ylabel('y')
    plt.title("Vertical displacement, $u_z$")
    bar = ax.imshow(uz, extent=[x_min+dx/2, x_max, y_min+dx/2, y_max], origin='lower',interpolation='bicubic', cmap=plt.cm.inferno)
    plt.colorbar(bar, label='Vertical displacement')
    # plot circle of radius a on the top
    theta = np.linspace(0, 2*np.pi, 100)
    plt.plot(x0 + a*np.cos(theta), y0 + a*np.sin(theta), linestyle="--", color='grey', zorder=2)
    # To save vertical displacement figure, uncomment this line
    # fig.savefig("Vertical_displacement_p0_{0:.3f}_a_{1:.3f}.png".format(p0,a), dpi=300)
    plt.show()

    # Plot in-plane displacement
    fig,ax = plt.subplots()
    ax.set_xlabel('x')
    ax.set_ylabel('y')
    plt.title("Inplane displacement, $u_x \\vec e_x + u_y \\vec e_y$")
    bar = ax.imshow(np.sqrt(ux**2+uy**2), extent=[x_min+dx/2, x_max, y_min+dx/2, y_max], origin='lower',interpolation='bicubic', cmap=plt.cm.inferno)
    plt.colorbar(bar, label='Displacement magnitude')
    if N > 30:
        every=N//30
    else:
        every=1
    print(np.max(ux))
    scale = 0.1/np.max(ux)
    plt.quiver(X[::every,::every], Y[::every,::every], scale*ux[::every,::every], scale*uy[::every,::every], color='black', scale=2,zorder=3)
    # plot circle of radius a on the top
    theta = np.linspace(0, 2*np.pi, 100)
    plt.plot(x0 + a*np.cos(theta), y0 + a*np.sin(theta), linestyle="--", color='grey', zorder=2)
    # To save in-plane displacement figure, uncomment this line
    # fig.savefig("Inplane_displacement_p0_{0:.3f}_a_{1:.3f}.png".format(p0,a), dpi=300)
    plt.show()

    #----------------------------------#
    # Show in-plane displacement in 3D #
    #----------------------------------#

    from mpl_toolkits.mplot3d import Axes3D
    fig,ax = plt.subplots(1,1,subplot_kw={'projection':'3d'})
    if N > 30:
        every=N//30
    else:
        every=1
    surf = ax.plot_surface(X, Y, np.sqrt(ux**2+uy**2), cmap=plt.cm.inferno, linewidth=0, antialiased=True, rstride=every, cstride=every)
    ax.set_xlabel('x')
    ax.set_ylabel('y')
    ax.set_zlabel('In-plane displacement')
    plt.show()

    #--------------------------------#
    # Show normal displacement in 3D #
    #--------------------------------#

    from mpl_toolkits.mplot3d import Axes3D
    fig,ax = plt.subplots(1,1,subplot_kw={'projection':'3d'})
    if N > 30:
        every=N//30
    else:
        every=1
    surf = ax.plot_surface(X, Y, uz, cmap=plt.cm.inferno, linewidth=0, antialiased=True, rstride=every, cstride=every)
    ax.set_xlabel('x')
    ax.set_ylabel('y')
    ax.set_zlabel('Vertical displacement')
    plt.show()
